/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "libraries/fstUtils.h"

using namespace fst;
using namespace std;

int getNumNodesFst(StdVectorFst* ifst) {
	int numnodes = 0;
	for (StateIterator< Fst<StdArc> > siter(*ifst);!siter.Done(); siter.Next())
		numnodes++;
	return numnodes;
}

int getFinalStateFst(StdVectorFst* ifst) {
	for (StateIterator< Fst<StdArc> > siter(*ifst);!siter.Done(); siter.Next()) {
		int st = siter.Value();
		if(ifst->Final(st) != StdFst::Weight::Zero())	//if final wt is non Zero(), then it is a final wt
			return st;
	}
	return -1;	//no finalstate
}

//this function does addeps ie introduce a new intial state and add eps trans to each orig state from it; also eps trans to final state from each orig state
void addEpsFst(StdVectorFst* newfst) {
	int n = getNumNodesFst(newfst); //assuming 0 is inital state
	newfst->AddState();
	newfst->AddState();
	for(int i=0; i<n; i++) {
		newfst->AddArc(i, StdArc(0, 0, 0, n+1, -1, -1)); //new eps arc has no prov info
		newfst->AddArc(n, StdArc(0, 0, 0, i, -1, -1)); //new eps arc has no prov info
	}
	newfst->SetStart(n);
	newfst->SetFinal(n+1,0);
}

//this subfst version also takes in the list of edges; includes only those edges 
StdVectorFst projectSubFst_edges(VectorFst<StdArc> *fst, int numnodes, int src_state, int sink_state, int* statepr, set < pair <int, int> > *edges) {
	if(src_state == -1)
		src_state = 0;
	if(sink_state == -1)
		sink_state == numnodes - 1;

	int mapst[numnodes];	//this maps from node num here to node num in new fst; in incr order
	int mapindex = 0;
	StdVectorFst outfst;
	for(int z=0; z<numnodes; z++) {
		if(statepr[z] == 1) {
			mapst[z] = mapindex;
			mapindex++;
			outfst.AddState();	//Add its corresp state
		}
		else
			mapst[z] = -1;	//this state z is not suppsd to be present in the subfst
	}
	outfst.SetStart(0);
	for (StateIterator< Fst<StdArc> > siter(*fst);!siter.Done(); siter.Next()) {
		int from = siter.Value();
		if(statepr[from] == 0)	//this state is not in the subfst, so ignore
			continue;
		for(ArcIterator< Fst<StdArc> > aiter(*fst, from);!aiter.Done(); aiter.Next()) {
			StdArc arc = aiter.Value();
			int to = arc.nextstate;
			if(statepr[to] == 0)	//this state is not present either
				continue;
			if(edges->find(make_pair(from, to)) == edges->end())	//this edge is not present in the given list
				continue;
			outfst.AddArc(mapst[from], StdArc(arc.ilabel, arc.olabel, arc.weight, mapst[to], arc.origfrom, arc.origto));
		}
	}
	//mapindex is now the number of states
	if(mapindex > 0)
		outfst.SetFinal((mapindex - 1), 0);
	else {
		outfst.AddState();
		outfst.SetFinal(0, 0);	//this is an empty fst; so, only one state
	}
	return outfst;
}

//obtain the min distances between every pair of nodes
void floydstuff(VectorFst<StdArc> *fst, int* nodedismin) {
		int numnodes = getNumNodesFst(fst);
		for(int i=0; i<numnodes; i++)
			for(int j=0; j<numnodes; j++)
				nodedismin[i*numnodes+j] = (i == j) ? 0 : INFIN;
		for (StateIterator< Fst<StdArc> > siter(*fst);!siter.Done(); siter.Next()) {
			int from = siter.Value();
			int nxtbit[numnodes];
			for(int nb=0; nb<numnodes; nb++)
				nxtbit[nb] = 0;
			for(ArcIterator< Fst<StdArc> > aiter(*fst, from);!aiter.Done(); aiter.Next()) {
				StdArc arc = aiter.Value();
				int to = arc.nextstate;
				if(nxtbit[to] == 1)	//state seen already in another arc of this transition
					continue;
				nxtbit[to] = 1;
				nodedismin[from*numnodes+to] = 1;
			}
		}
		//Floyd Warshall Algo
		for(int k=0; k<numnodes; k++) {
			for(int i=0; i<numnodes; i++) {
				for(int j=0; j<numnodes; j++) {
					int mini = nodedismin[i*numnodes+j];
					int sum = nodedismin[i*numnodes+k] + nodedismin[k*numnodes+j];
					nodedismin[i*numnodes+j] = mini > sum ? sum : mini;
				}
			}
		}
}

//this function is eqvltnt to fst::write, converting an in-mem fst to mtxt file
void fsttombin(StdVectorFst* fst, ostream &fout) {
	int numnodes = getNumNodesFst(fst);
	int matr[numnodes][numnodes];	//This is the adjacency matrix
	for(int i=0; i<numnodes; i++) {
		for(int j=0; j<numnodes; j++) {
			matr[i][j] = 0;
		}
	}
	int startstate = 0;	//assuming the input fst is toposorted and starts with state 0
	int finalstate = getFinalStateFst(fst);
	for (StateIterator< Fst<StdArc> > siter(*fst); !siter.Done(); siter.Next()) {
		int st = siter.Value();	//from state
		for(ArcIterator< Fst<StdArc> > aiter(*fst, st);!aiter.Done(); aiter.Next()) {
			StdArc arc = aiter.Value();
			int to = arc.nextstate;
			matr[st][to] = 1;
		}
	}
	fout.write((char*)&numnodes, sizeof(numnodes));
	fout.write((char*)&startstate, sizeof(startstate));
	fout.write((char*)&finalstate, sizeof(finalstate));
	//now check the adj matrx for each state, and pick those arcs leaving this state
	int numtrans = 0;
	for(int i=0; i<numnodes; i++)
		for(int j=0; j<numnodes; j++)
			if(matr[i][j] == 1)
				numtrans++;
	fout.write((char*)&numtrans, sizeof(numtrans));
	for (StateIterator< Fst<StdArc> > siter(*fst);!siter.Done(); siter.Next()) {
		int st = siter.Value();	//from state
		vector<int> nextnodes;
		for(int i=0; i<numnodes; i++) {
			if(matr[st][i] == 1)
				nextnodes.push_back(i);
		}
		int numnextnodes = nextnodes.size();
		for(int i=0; i < numnextnodes; i++) {	//for each next node, get its arcs
			fout.write((char*)&st, sizeof(st));
			fout.write((char*)&nextnodes[i], sizeof(nextnodes[i]));
			int numarcs = 0;
			for(ArcIterator< Fst<StdArc> > aiter(*fst, st);!aiter.Done(); aiter.Next()) {
				StdArc arc = aiter.Value();
				if(arc.nextstate == nextnodes[i])
					numarcs++;
			}
			fout.write((char*)&numarcs, sizeof(numarcs));
			for(ArcIterator< Fst<StdArc> > aiter(*fst, st);!aiter.Done(); aiter.Next()) {
				StdArc arc = aiter.Value();
				if(arc.nextstate == nextnodes[i]) {
					numarcs++;				
					int ilabel = arc.ilabel;
					int olabel = arc.olabel;
					double arcwt = arc.weight.Value();
					fout.write((char*)&ilabel, sizeof(ilabel));
					fout.write((char*)&olabel, sizeof(olabel));
					fout.write((char*)&arcwt, sizeof(arcwt));
				}
			}
		}
	}
}


//this function is eqvltnt to fst::read, converting a stringstream to an in-mem fst
void mbintofst(StdVectorFst *ifst, char* buffer) {
	int numst, sst, est, numtrans;
	int sizeofint = sizeof(int);
	int sizeofdouble = sizeof(double);
	memcpy((char*)&numst, buffer, sizeofint);
	memcpy((char*)&sst, buffer+sizeofint, sizeofint);
	memcpy((char*)&est, buffer+2*+sizeofint, sizeofint);
	memcpy((char*)&numtrans, buffer+3*+sizeofint, sizeofint);
	int addendum = 4*sizeofint;	//to be added to buffer
	for(int i=0; i<numst; i++)
		ifst->AddState();
	ifst->SetStart(sst);
	ifst->SetFinal(est, 0);
	for(int nt=0; nt<numtrans; nt++) {	//for each transn
		int frst, tost, numarcs;
		memcpy((char*)&frst, buffer+addendum, sizeofint);
		memcpy((char*)&tost, buffer+addendum+sizeofint, sizeofint);
		memcpy((char*)&numarcs, buffer+addendum+2*sizeofint, sizeofint);
		addendum += 3*sizeofint; 
		for(int na=0; na<numarcs; na++) {
			int ilabel, olabel;
			double weight;
			memcpy((char*)&ilabel, buffer+addendum, sizeofint);
			memcpy((char*)&olabel, buffer+addendum+sizeofint, sizeofint);
			memcpy((char*)&weight, buffer+addendum+2*sizeofint, sizeofdouble);
			addendum += 2*sizeofint + sizeofdouble;
			ifst->AddArc(frst, StdArc(ilabel, olabel, weight, tost, frst, tost));	//provenance info included
		}
	}
}

//this function is eqvltnt to fst::read, converting a stringstream to an in-mem fst
void mbintofst(StdVectorFst *ifst, istream &fststr) {
	fststr.seekg (0, iostream::end);
	int b1size = fststr.tellg();
	fststr.seekg(0, iostream::beg);  //reposition back to beginning
	char buf[b1size];
	memset(buf, '\0', b1size);
	fststr.read(buf, b1size);
	mbintofst(ifst, buf);
}

//this function augments the arcs in the fst with the origfrom and origto info on them and copies
//this assumes that the ifst has single start and end states, and has last state num as end state
//in addition, this function changes the arc weights based on the normalization of probs, ie sum of leaving arcprobs is 1
StdVectorFst augmentFst(StdVectorFst* ifst) {
	int numnodes = getNumNodesFst(ifst);
	vector <double> outprobs[numnodes];	//the list of probs outgoing from a node; later aggr these probs
	for (StateIterator< Fst<StdArc> > siter(*ifst);!siter.Done(); siter.Next()) {
		int stfrom = siter.Value();
		for(ArcIterator< Fst<StdArc> > aiter(*ifst, stfrom);!aiter.Done(); aiter.Next()) {
			StdArc arc = aiter.Value();
			outprobs[stfrom].push_back(arc.weight.Value());
		}
	}
	//compute the net outgoing prob for each node (and thus, the normalizn constants) for each node
	//since sum(e^-wti) = p, the normalzn factor is 1/p, which means prob is 1/p*e^-wti; thus new wt is just wti + lnp
	double wtnormlzn[numnodes];	//lnp factor for each node
	//space-only trans are handled separately; the assmptn is that spaceonly transns are the lone transn out of that node; so check only the frnode
	set <int> spaceonlies;
	for(int i=0; i<numnodes; i++) {
		double netprob = 0.0;
		int numouts = outprobs[i].size();
		for(int op=0; op < numouts; op++) {
			netprob += exp(-1*outprobs[i][op]);
		}
		if(netprob > 20)//to account for space only transns! (94*e^-1 = 34.5, so it is captured)
			spaceonlies.insert(i);
		double wtnml = log(netprob);
		wtnormlzn[i] = wtnml;
	}
	//create new fst with provenance info and normalized prob wts; fst is topsorted
	StdVectorFst ofst;
	int numst = getNumNodesFst(ifst);
	for(int i=0; i<numst; i++)
		ofst.AddState();
	ofst.SetStart(0);
 	for (StateIterator< Fst<StdArc> > siter(*ifst);!siter.Done(); siter.Next()) {
		int st = siter.Value();
		//check if it is a spaceonly; if so, add special arcs to the next transn, if not, just normalize
		//the assmpn is that only one transn is present, and hence any arc can give the nextstate
		set<int>::iterator findst = spaceonlies.find(st);
		if(findst != spaceonlies.end()) {
			ArcIterator< Fst<StdArc> > aiter(*ifst, st);
			StdArc arc = aiter.Value();
			int nextst = arc.nextstate;
			ofst.AddArc(st, StdArc(arc.ilabel, 32, 0, nextst, st, nextst));			//100% certain space
			
			//assumption made that 126 characters are present for the olabel (so, only OCR)
			//NOTE: this function is no longer portable for other fsts with diff osyms (incl. speech)
			double restwt = 100;				//0 prob
			for(int ola = 0; ola <= 125; ola++) {
				if(ola == 32) continue;
				ofst.AddArc(st, StdArc(arc.ilabel, ola, restwt, nextst, st, nextst));	//0 prob
			}
		}
		else {		
			for(ArcIterator< Fst<StdArc> > aiter(*ifst, st);!aiter.Done(); aiter.Next()) {
				StdArc arc = aiter.Value();
				double newwt = (double)arc.weight.Value() + wtnormlzn[st];
				ofst.AddArc(st, StdArc(arc.ilabel, arc.olabel, newwt, arc.nextstate, st, arc.nextstate)); //include the origfrom and origto info for this new arc
			}
		}
	}
	ofst.SetFinal(numst-1,0);
	return ofst;
}

//this function computes the reachability probabilities by dyn progmg
vector<double> dynReachProbs (StdVectorFst *ifst) {
	vector<double> retreachprobs;
	int numnodes = getNumNodesFst(ifst);	
	map<int, vector <double> > parindivprobs[numnodes];	//the list of probs per parent; later aggr these probs
	for (StateIterator< Fst<StdArc> > siter(*ifst);!siter.Done(); siter.Next()) {
		int stfrom = siter.Value();
		for(ArcIterator< Fst<StdArc> > aiter(*ifst, stfrom);!aiter.Done(); aiter.Next()) {
			StdArc arc = aiter.Value();
			int stto = arc.nextstate;
			parindivprobs[stto][stfrom].push_back(arc.weight.Value());
		}
	}
	map<int, double> parprobs[numnodes]; //compute the par to child prob by aggr arcs
	for(int i=0; i<numnodes; i++) {
		for(map<int, vector <double> >::iterator iter = parindivprobs[i].begin(); iter != parindivprobs[i].end(); iter++) {
			double netprob = 0.0;
			int numprobs = iter->second.size();
			for(int ar=0; ar<numprobs; ar++)
				netprob += exp(-1*iter->second[ar]);
			parprobs[i][iter->first] = netprob;
			if(netprob > 2)
				parprobs[i][iter->first] = 1.0;	//reseting for space-only xns
		}
		
	}
	//fst is in topsort order; start from 0 and go to last node, and chk each's parents
	double zeroprob;
	zeroprob = 1;	//start node's prob is 1
	retreachprobs.push_back(zeroprob);
	for(int i=1; i<numnodes; i++) {
		double thisprob = 0.0;
		for(map<int, double>::iterator iter = parprobs[i].begin(); iter != parprobs[i].end(); iter++) {
			double thisparprob = retreachprobs[iter->first]; //chk this parent's prob; mult that by transn prob, and add to existing prob
			double transprob = parprobs[i][iter->first];
			thisprob += thisparprob * transprob;	//thisprob is sum over parents of parprob * transprob
		}
		retreachprobs.push_back(thisprob);
	}
	return retreachprobs;
}

void getQueryMatrix(StdVectorFst* query, int* dfaquery) {
	int NUMST = getNumNodesFst(query);
	for(int i=0; i<NUMST; i++)
		for(int j=0; j<=125; j++)
			dfaquery[i*126 + (j)]= -1;	//-1 means no state mentioned	-> this shouldnt happen in the final dfa matrix
	for (StateIterator< Fst<StdArc> > siter(*query);!siter.Done(); siter.Next()) {
		int from = siter.Value();
		for(ArcIterator< Fst<StdArc> > aiter(*query, from);!aiter.Done(); aiter.Next()) {
			StdArc arc = aiter.Value();
			int to = arc.nextstate;
			char label = arc.ilabel;
			dfaquery[from*126 + (label - '\0')] = to;		//next state
		}
	}
	int finalstate = getFinalStateFst(query);	//get the final state of given fst
	for(int j=0; j<=125; j++)
		dfaquery[finalstate*126 + (j)]= finalstate;	//finalstate has a .* loop added onto it -> by resetting previous transitions; it is accpeted anyways
}

void constructIterator (PathIterator* incrementalN, StdVectorFst* ifst) {
	PathEnumerator	*enumeratorN;
	const SymbolTable *isyms = 0, *osyms = 0, *ssyms = 0;
	isyms = ifst->InputSymbols();
	osyms = ifst->OutputSymbols();			
	enumeratorN = new PathEnumerator(*ifst, isyms, osyms, ssyms, false,false);
	enumeratorN->fstToGraph();
	incrementalN->moreflag = 1;
	incrementalN->setPathEnumerator(enumeratorN);	
}

void constructIteratorNoFst (PathIterator* incrementalN, blockskel thisskel, map <Edge, block> *blockdata, int* scratch) {
	PathEnumerator	*enumeratorN;
	StdVectorFst *ifst;	
	const SymbolTable *isyms = 0, *osyms = 0, *ssyms = 0;
	enumeratorN = new PathEnumerator(*ifst, isyms, osyms, ssyms);
	enumeratorN->buildGraph(thisskel, blockdata, scratch);
	incrementalN->moreflag = 1;
	incrementalN->setPathEnumerator(enumeratorN);	
}
